/**
 * Copyright (c) 2008-2021, OPPO Mobile Comm Corp., Ltd
 * VENDOR_EDIT
 * @file hey_schema_api.h
 * @brief 设备抽象模型操作接口
 * @version 0.1
 * @date 2021-01-27
 * @author WatWu
 */

#ifndef HEY_SCHEMA_API_H
#define HEY_SCHEMA_API_H

#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>

struct hey_service;
struct hey_action_ctx;

/**
 * @brief 设备属性被动操作回调函数列表
 */
struct hey_properties_callbacks {
    /**
    * @brief 对端对设备属性进行写入
    * @param cb_data    [服务注册的回调数据]
    * @param n_iid      [属性id数量]
    * @param iids       [属性id列表]
    * @param len        [写入属性的数据长度]
    * @param value      [写入属性的原始数据]
    * @return int       [回调处理结果]
    */
    int (*write)(void *cb_data, size_t n_iid, unsigned int *iids,
                 size_t len, unsigned char *value);

    /**
    * @brief 对端对设备属性数组添加元素
    * @param cb_data    [服务注册的回调数据]
    * @param iid        [被添加元素的数组属性id]
    * @param len        [添加的属性元素数据长度]
    * @param value      [添加的属性元素原始数据]
    * @return int       [添加的元素在数组中的id，负值为错误码]
    */
    int (*array_add)(void *cb_data, unsigned int iid,
                     size_t len, unsigned char *value);

    /**
    * @brief 对端对设备属性数组删除元素
    * @param cb_data    [服务注册的回调数据]
    * @param iid        [被删除元素的数组属性id]
    * @param n_id       [被删除元素数量]
    * @param ids        [被删除元素在数组中的id列表]
    * @return int       [回调处理结果]
    */
    int (*array_del)(void *cb_data, unsigned int iid, size_t n_id, unsigned int *ids);

    /**
    * @brief 对端对设备属性数组更新元素
    * @param cb_data    [服务注册的回调数据]
    * @param iid        [被更新元素的数组属性id]
    * @param id         [被更新元素在数组中的id]
    * @param len        [更新的属性元素数据长度]
    * @param value      [更新的属性元素原始数据]
    * @return int       [回调处理结果]
    */
    int (*array_replace)(void *cb_data, unsigned int iid, unsigned int id,
                         size_t len, unsigned char *value);
};

/**
 * @brief 动作类型定义
 */
enum hey_action_type {
    HEY_ACTION_ONCE = 0,
    HEY_ACTION_IN_ONCE_OUT_STREAM = 1,
    HEY_ACTION_IN_STREAM_OUT_ONCE = 2,
    HEY_ACTION_IN_STREAM_OUT_STREAM = 3
};

/**
 * 流式动作上下文回调函数列表
 */
struct hey_action_ctx_callbacks {
    /**
    * @brief 对端对设备发起动作交互
    * @param cb_data        [服务注册的回调数据]
    * @param ctx            [动作的上下文，SDK生成，应用层负责保存及释放]
    * @param len            [动作数据长度]
    * @param value          [动作原始数据]
    * @param user_context   [自定义上下文，SDK回调时带回，应用层生成及负责释放]
    * @return int           [回调处理结果]
    */
    int (*start)(void *cb_data, struct hey_action_ctx *ctx, size_t len, unsigned char *value,
                 void **user_context);

    /**
    * @brief 流式动作交互的消息接收回调
    * @param cb_data        [服务注册的回调数据]
    * @param user_context   [start回调中传入的上下文]
    * @param len            [动作数据长度]
    * @param value          [动作原始数据]
    * @return int           [回调处理结果]
    */
    int (*message)(void *cb_data, void *user_context, size_t len, unsigned char *value);

    /**
    * @brief 对端对设备停止动作交互
    * @param cb_data        [服务注册的回调数据]
    * @param user_context   [start回调中传入的上下文]
    */
    void (*stop)(void *cb_data, void *user_context);
};

/**
 * @brief 设备抽象动作定义入口
 */
struct hey_action_entry {
    unsigned int iid;
    enum hey_action_type type;
    const struct hey_action_ctx_callbacks *cbs; // 函数表应使用静态变量定义
};

/**
 * @brief 动作交互的消息发送接口
 * @param ctx       [动作的上下文]
 * @param len       [动作数据长度]
 * @param value     [动作原始数据]
 * @return int      [动作消息发送结果，表示动作消息数据是否成功交付到SDK中]
 */
int hey_action_msg_send(struct hey_action_ctx *ctx,
                        size_t len, unsigned char *value);

/**
 * @brief 主动停止动作交互
 * @param ctx       [动作的上下文]
 * @return int      [操作结果]
 */
int hey_action_stop(struct hey_action_ctx *ctx);

/**
 * @brief 设备抽象属性
 */
struct hey_schema_properties {
    size_t n_iid;                               // 属性ID数量
    unsigned int *iids;                         // 属性ID列表
    const struct hey_properties_callbacks *cbs; // 函数表应使用静态变量定义
    size_t len;                                 // 属性数据长度
    unsigned char *value;                       // 属性数据内容
};

/**
 * @brief 设备抽象动作
 */
struct hey_schema_actions {
    size_t n_iid;                  // 动作ID数量
    struct hey_action_entry *iids; // 动作结构列表
};

/**
 * @brief 设备抽象事件
 */
struct hey_schema_events {
    size_t n_iid;       // 事件ID数量
    unsigned int *iids; // 事件ID列表
};

/**
 * @brief 设备抽象服务，包含属性、动作、事件
 */
struct hey_schema_service {
    struct hey_schema_properties properties;
    struct hey_schema_actions actions;
    struct hey_schema_events events;
};

/**
 * @brief initialize hey schema service
 * @param service [schema service]
 */
void hey_schema_service_init(struct hey_schema_service *service);

/**
 * @brief 向SDK注册一个设备抽象服务，由SDK负责服务属性等的内部交互管理
 * @param siid          [待注册的服务编号]
 * @param service       [待注册的抽象服务]
 * @param cb_data       [注册回调数据]
 * @return hey_service  [已注册的服务对象，NULL则为注册失败]
 */
struct hey_service *hey_service_register(unsigned int siid, struct hey_schema_service *service,
                                         void *cb_data);

/**
 * @brief 属性更新的内容
 */
struct hey_properties_update_content {
    size_t n_iid;
    unsigned int *iids;
    size_t len;
    unsigned char *value;
};

/**
 * @brief 设备属性改变更新到SDK
 * @param service   [服务对象]
 * @param content   [属性更新内容]
 * @return int      [更新到SDK的结果]
 */
int hey_properties_update(struct hey_service *service,
                          struct hey_properties_update_content *content);

/**
 * @brief 事件通知的内容
 */
struct hey_event_notify_content {
    unsigned int iid;        // 事件ID
    unsigned int importance; // 事件优先级
    uint64_t ref_id;         // 事件通知关联的UUID
    uint64_t timestamp;
    size_t len;
    unsigned char *value;
    void *cb_data;
    void (*notify_result_cb)(void *cb_data, int result);
};

/**
 * @brief 事件通知
 * @param service   [服务对象]
 * @param content   [实际通知的内容]
 * @return int64_t  [事件通知id，负值为错误码]
 */
int64_t hey_event_notify(struct hey_service *service,
                         struct hey_event_notify_content *content);

#endif
