/**
 * Copyright (c) 2008-2021, OPPO Mobile Comm Corp., Ltd
 * VENDOR_EDIT
 * @file heythings.h
 * @brief heythings sdk api
 * @version 0.1
 * @date 2021-03-04
 * @author WatWu
 */

#ifndef HEYTHINGS_SDK_API_H
#define HEYTHINGS_SDK_API_H

#include <stdbool.h>
#include <stdint.h>
#include <unistd.h>
#include <stdio.h>

#define HEY_BRAND_LEN 32
#define HEY_MANUF_LEN 32
#define HEY_MODEL_LEN 32
#define HEY_HARDVER_LEN 32
#define HEY_SOFTVER_LEN 32

enum hey_sdk_state {
    HEY_SDK_STATE_INITIAL = 0,
    HEY_SDK_STATE_WIFI_CONNECTING = 1,
    HEY_SDK_STATE_SERVER_CONNECTING = 2,
    HEY_SDK_STATE_SERVER_CONNECTED = 3
};

enum hey_setup_state {
    HEY_SETUP_STATE_INITIAL = 0,
    HEY_SETUP_STATE_RUNNING = 1,
    HEY_SETUP_STATE_FAILED = 2,
    HEY_SETUP_STATE_SUCCESS = 3
};

enum hey_setup_confirm_opinion {
    HEY_SETUP_ACCEPT = 1,
    HEY_SETUP_REJECT = 2
};

enum hey_log_level {
    HEY_LOG_LEVEL_ERR = 3,
    HEY_LOG_LEVEL_WARN = 4,
    HEY_LOG_LEVEL_NITICE = 5,
    HEY_LOG_LEVEL_INFO = 6,
    HEY_LOG_LEVEL_DEBUG = 7
};

struct hey_dev_info_config {
    unsigned char brand[HEY_BRAND_LEN + 1];
    unsigned char manufacture[HEY_MANUF_LEN + 1];
    unsigned char model[HEY_MODEL_LEN + 1];
    unsigned char hard_version[HEY_HARDVER_LEN + 1];
    unsigned char soft_version[HEY_SOFTVER_LEN + 1];
};

struct hey_sdk_callbacks {
    void (*state_callback)(enum hey_sdk_state state);
    void (*time_synced)(void); // SDK联网后时间同步完成，应用可使用标准C接口获取时间
    void (*setup_reset)(void); // 配网功能复位回调
};

struct hey_sdk_config {
    struct hey_dev_info_config info;
    const struct hey_sdk_callbacks *cbs;
};

struct hey_setup_callbacks {
    void (*provision_confirm)(void);                    // 请求配网授权回调
    void (*set_pin)(char **pin, size_t *pin_len);       // 设置配网交互过程的PIN
    void (*state_callback)(enum hey_setup_state state); // 配网状态回调
};

struct hey_setup_cert_key {
    uint8_t *dev_cert;           // 设备证书
    size_t dev_cert_len;         // 设备证书长度
    uint8_t *dev_private_key;    // 设备私钥
    size_t dev_private_key_len;  // 设备私钥长度
    uint8_t *product_cert;       // 产品证书
    size_t product_cert_len;     // 产品证书长度
    uint8_t *ecdh_private_key;   // ECDH私钥，仅使用ECDH配网的产品才有
    size_t ecdh_private_key_len; // ECDH私钥长度
};

struct hey_setup_config {
    uint32_t timeout_seconds;
    struct hey_setup_cert_key cert_key;
    const struct hey_setup_callbacks *cbs;
};

/**
 * @brief heythings sdk initialization
 * @param conf [sdk configuration]
 * @return int [initialized result]
 */
int hey_sdk_init(struct hey_sdk_config *conf);

/**
 * @brief heythings sdk starts to run, it will not return if it starts successfully.
 * @return int [sdk runs result, not 0 means sdk runs failed]
 */
int hey_sdk_run(void);

/**
 * @brief start a network setup process, only one setup can run at the same time.
 * @param conf  [network setup parameters]
 * @return int  [start result，0 means success]
 */
int hey_setup_start(struct hey_setup_config *conf);

/**
 * @brief confirms the setup provision to the SDK
 * @param opinion [confirm opinion]
 */
void hey_setup_confirm(enum hey_setup_confirm_opinion opinion);

/**
 * @brief stop a running network setup process or reset the completed setup
 * @return int [reset result，0 means success]
 */
int hey_setup_reset(void);

/**
 * @brief get setup state.
 * @return enum [setup state]
 */
enum hey_setup_state hey_setup_get_state(void);

/**
 * @brief Set the log enable
 * @param enable [log enable, true or false]
 */
void hey_log_set_enable(bool enable);

/**
 * @brief Get the log enable flag
 * @return [log enable, true or false]
 */
bool hey_log_get_enable(void);

/**
 * @brief start to upload log
 */
void hey_log_upload(void);

/**
 * @brief : if log enable, it will be printf to sdk internal log buffer at the same time.
 * @param level     [log level]
 * @param format    [log strings to be formatted]
 * @param ...       [format content]
 */
void hey_log_printf(enum hey_log_level level, const char *format, ...);

/**
 * @brief The callback function of the peer reveived array add method command.
 */
typedef void (*array_add_received_cb_t)(void *cb_data);

/**
 * @brief send heythings array add method
 * @param cb_data   [callback data]
 * @param cb_func   [callback function of the peer reveived method]
 * @param siid      [service id]
 * @param iid       [property id]
 * @param data      [array data]
 * @param data_len  [data length]
 * @return int      [operation result]
 */
int hey_send_array_add_req(void *cb_data, array_add_received_cb_t cb_func,
                           uint32_t siid, uint32_t iid, void *data, size_t data_len);

#endif
