#include "cli_task.h"
#include "driver/uart.h"
#include "heythings.h"
#include <string.h>
#include "user_task.h"
#include "schema/light/light_service.h"

#define BUF_SIZE (512)

#define DEBUG_UART_PORT UART_NUM_0

struct command {
    const char *name;
    void (*handle)(void);
    const char *description;
};

static void help(void);
static void reset(void);
static void setup(void);
static void get_sdk_state(void);
static void reboot(void);
static void accept_setup(void);
static void turn_on_light(void);
static void turn_off_light(void);

static struct command s_commands[] = {
    { "help", help, "" },
    { "reset", reset, "unbind device" },
    { "setup", setup, "device enter the binding process" },
    { "reboot", reboot, "reboot system" },
    { "state", get_sdk_state, "printf states, including mem cpu etc." },
    { "accept", accept_setup, "confirm setup to continue" },
    { "on", turn_on_light, "turn on light" },
    { "off", turn_off_light, "turn off light" },
};

static void help(void)
{
    printf("command list:\r\n");
    for (int i = 1; i < sizeof(s_commands) / sizeof(struct command); i++) {
        printf("\t%s:\t%s\r\n", s_commands[i].name, s_commands[i].description);
    }
}

static void reset(void)
{
    hey_setup_reset();
    printf("run hey_setup_reset()\r\n");
    esp_restart();
}

static void setup(void)
{
    device_start_to_setup();
    printf("run hey_setup_start()\r\n");
}

/**
 * @brief 直接串口打印系统任务管理器
 */
static void printf_os_task_manager(void)
{
    char *pcWriteBuffer = malloc(1024);
    int free_mem = 0;

    printf("\n=============================================\n");
    printf("name \t\tstatus \tprio \tfree \tpid\n");
    vTaskList((char *)pcWriteBuffer);
    printf("%s", pcWriteBuffer);
    printf("---------------------------------------------\n");
    printf("name \t\trun_count \tusage\n");
    vTaskGetRunTimeStats((char *)pcWriteBuffer);
    printf("%s", pcWriteBuffer);
    printf("---------------------------------------------\n");
    free_mem = esp_get_free_heap_size();
    printf("memory free : \t%d byte\n", free_mem);
    printf("=============================================\n\n");

    free(pcWriteBuffer);
}

static void get_sdk_state(void)
{
    printf("device bind state : ");
    switch (hey_setup_get_state()) {
    case HEY_SETUP_STATE_SUCCESS:
        printf("bound\r\n");
        break;
    default:
        printf("unbound\r\n");
        break;
    }
    printf_os_task_manager();
}

static void reboot(void)
{
    esp_restart();
}

static void accept_setup(void)
{
    hey_setup_confirm(HEY_SETUP_ACCEPT);
}

static void turn_on_light(void)
{
    struct light_properties props = { 0 };
    props.power = true;
    unsigned int enumerate[1] = { LIGHT_PROPERTY_POWER };
    int ret = light_properties_update(1, enumerate, &props);
    if (ret != 0) {
        printf("propeties update failed : %d\n", ret);
    }
}

static void turn_off_light(void)
{
    struct light_properties props = { 0 };
    props.power = false;
    unsigned int enumerate[1] = { LIGHT_PROPERTY_POWER };
    int ret = light_properties_update(1, enumerate, &props);
    if (ret != 0) {
        printf("propeties update failed : %d\n", ret);
    }
}

static struct command *find(char *cmd_str, struct command *s_commands, size_t cmds_len)
{
    int i = 0;
    for (i = 0; i < cmds_len; i++) {
        if (strcmp(s_commands[i].name, cmd_str) == 0) {
            return &s_commands[i];
        }
    }

    return NULL;
}

static void process_line(char *cmd_str, size_t cmd_len)
{
    struct command *command = { 0 };
    command = find(cmd_str, s_commands, sizeof(s_commands) / sizeof(struct command));
    if (command && command->handle) {
        command->handle();
    }
}

void cli_task(void *pvParameter)
{
    uart_config_t uart_config = {
        .baud_rate = CONFIG_CONSOLE_UART_BAUDRATE,
        .data_bits = UART_DATA_8_BITS,
        .parity = UART_PARITY_DISABLE,
        .stop_bits = UART_STOP_BITS_1,
        .flow_ctrl = UART_HW_FLOWCTRL_DISABLE
    };
    uart_param_config(DEBUG_UART_PORT, &uart_config);
    uart_driver_install(DEBUG_UART_PORT, BUF_SIZE * 2, 0, 0, NULL, 0);

    // Configure a temporary buffer for the incoming data
    unsigned char *data = (unsigned char *)malloc(BUF_SIZE);

    int index = 0;

    while (1) {
        if (index > BUF_SIZE) {
            index = 0;

            hey_log_printf(HEY_LOG_LEVEL_INFO, "cli:cmd too long. not suport it.");
            continue;
        }
        // Read data from the UART
        int len = uart_read_bytes(DEBUG_UART_PORT, data + index, BUF_SIZE, 20 / portTICK_RATE_MS);

        if (len <= 0) {
            vTaskDelay(10);
            continue;
        }

        // Write data back to the UART
        uart_write_bytes(DEBUG_UART_PORT, (const char *)data + index, len);

        index += len;
        size_t buf_length = index < BUF_SIZE ? index : BUF_SIZE;

        for (int i = 0; i < buf_length; i++) {
            switch (data[i]) {
            case '\r':
            case '\n':
                if (i < 2) {
                    uart_write_bytes(DEBUG_UART_PORT, "\r\n> ", 4);
                } else {
                    data[i] = '\0';
                    process_line((char *)data, i + 1);
                }
                i = buf_length;
                index = 0;
                break;
            // case 0x03: // ASCII for Ctrl-C
            //     break;
            // case 0x04: // ASCII for Ctrl-D
            //     break;
            default:
                break;
            }
        }
    }
    free(data);
    vTaskDelete(NULL);
}
