/**
 * Copyright (c) 2008-2019, OPPO Mobile Comm Corp., Ltd
 * VENDOR_EDIT
 * @file ota_task.c
 * @brief ota task on esp32
 * @version 0.1
 * @date 2019-08-26
 * @author WatWu
 */

#include <string.h>
#include "mbedtls/ecdsa.h"
#include "mbedtls/pk.h"

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_log.h"
#include "esp_ota_ops.h"
#include "esp_https_ota.h"

#include "schema/softwareUpdate/software_update_service.h"
#include "heythings.h"
#include "app_version.h"

#define OTA_URL_SIZE 512

static const char *TAG = "ota_task";

static char s_ota_url[OTA_URL_SIZE] = { 0 };
static uint32_t total_len = 0;
static uint32_t total_len_of_percent = 0;
static uint32_t current_len = 0;
static uint8_t curr_ota_percent = 0;
static uint8_t last_ota_percent = 0;
static uint8_t ota_is_running = 0;

static mbedtls_md_context_t md_ctx;
static unsigned char *s_sign_data = NULL;
static int s_sign_len = 0;
static bool s_sign_flag = false;

/*****************************************************************************
 * This public key is only used for testing, do not use in actual products ! *
 ****************************************************************************/
const char *ota_sign_pubkey_pem = "-----BEGIN PUBLIC KEY-----\r\n"
                                  "MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE0P59bu8g6vAYOx/C5Nd4aL+fcmRf\r\n"
                                  "e021i5zqPQFSFcq1aKhsGmh3ndI5Q0nSNzmQYTV10lhJsBfzxzp3TDRoIA==\r\n"
                                  "-----END PUBLIC KEY-----\r\n";

/**
 * @brief verify ota bin with ecdsa signature.
 * @param char [hash of ota bin]
 * @param hlen [hash length]
 * @param char [signature data, ANS.1 format]
 * @param slen [signature length]
 * @return int [verify result, 0 means success]
 */
int verify_ota_bin(const unsigned char *hash, size_t hlen, const unsigned char *sig, size_t slen)
{
    int ret;
    mbedtls_ecdsa_context ctx;
    mbedtls_pk_context public_key;

    mbedtls_ecdsa_init(&ctx);
    mbedtls_pk_init(&public_key);

    //secp256r1
    ret = mbedtls_ecp_group_load(&ctx.grp, MBEDTLS_ECP_DP_SECP256R1);
    if (ret != 0) {
        hey_log_printf(HEY_LOG_LEVEL_INFO, "Failed to load secp256r1, error : -0x%04X.", -ret);
        goto cleanup;
    }

    //加载公钥
    ret = mbedtls_pk_parse_public_key(&public_key, (unsigned char *)ota_sign_pubkey_pem, strlen(ota_sign_pubkey_pem) + 1);
    if (ret != 0) {
        hey_log_printf(HEY_LOG_LEVEL_INFO, "Failed to extrace public key from pem format, error : -0x%04X.", -ret);
        goto cleanup;
    }
    mbedtls_ecp_keypair *p_ecp_keypair = mbedtls_pk_ec(public_key);
    mbedtls_ecp_copy(&ctx.Q, &p_ecp_keypair->Q);

    //ECDSA验签
    ret = mbedtls_ecdsa_read_signature(&ctx, hash, hlen, sig, slen);
    if (ret != 0) {
        hey_log_printf(HEY_LOG_LEVEL_INFO, "Failed to verify signature, error : -0x%04X.", -ret);
        goto cleanup;
    }

cleanup:
    mbedtls_ecdsa_free(&ctx);
    mbedtls_pk_free(&public_key);

    return (ret != 0);
}

static int report_device_soft_update_property_change(uint32_t progress, uint32_t state, uint32_t error_code, char *version)
{
    struct software_update_properties props;
    props.currentsoftwareversion = version;
    props.updateerror = error_code;
    props.updateprogress = progress;
    props.updatestate = state;

    uint32_t iids[4] = { SOFTWARE_UPDATE_PROPERTY_CURRENTSOFTWAREVERSION,
                         SOFTWARE_UPDATE_PROPERTY_UPDATEPROGRESS,
                         SOFTWARE_UPDATE_PROPERTY_UPDATESTATE,
                         SOFTWARE_UPDATE_PROPERTY_UPDATEERROR };
    software_update_properties_update(4, iids, &props);
    return 0;
}

static esp_err_t http_event_handler(esp_http_client_event_t *evt)
{
    switch (evt->event_id) {
    case HTTP_EVENT_ERROR:
        ESP_LOGE(TAG, "HTTP_EVENT_ERROR");
        break;
    case HTTP_EVENT_ON_CONNECTED:
        ESP_LOGI(TAG, "HTTP_EVENT_ON_CONNECTED");
        break;
    case HTTP_EVENT_HEADER_SENT:
        ESP_LOGI(TAG, "HTTP_EVENT_HEADER_SENT");
        break;
    case HTTP_EVENT_ON_HEADER:
        ESP_LOGI(TAG, "HTTP_EVENT_ON_HEADER, key=%s, value=%s", evt->header_key, evt->header_value);
        if (strcmp("Content-Length", evt->header_key) == 0) {
            total_len = atoi(evt->header_value);
            total_len_of_percent = total_len / 100;
            ESP_LOGI(TAG, "The total length of bin file is : %d", total_len);
        }
        break;
    case HTTP_EVENT_ON_DATA:
        ESP_LOGD(TAG, "HTTP_EVENT_ON_DATA, len=%d", evt->data_len);
        if (s_sign_flag) {
            mbedtls_md_update(&md_ctx, evt->data, evt->data_len);
        }
        current_len += evt->data_len;
        curr_ota_percent = current_len / total_len_of_percent;
        if (curr_ota_percent != last_ota_percent) {
            last_ota_percent = curr_ota_percent;
            ESP_LOGI(TAG, "Upgrade progress : %%%d !", last_ota_percent);
            if (last_ota_percent % 5 == 0) {
                report_device_soft_update_property_change(last_ota_percent,
                                                          SOFTWARE_UPDATE_UPDATE_STATE_DOWNLOADING,
                                                          SOFTWARE_UPDATE_UPDATE_ERROR_INITIAL,
                                                          FIRMWARE_VERSION);
            }
        }
        break;
    case HTTP_EVENT_ON_FINISH:
        ESP_LOGI(TAG, "HTTP_EVENT_ON_FINISH");
        break;
    case HTTP_EVENT_DISCONNECTED:
        ESP_LOGE(TAG, "HTTP_EVENT_DISCONNECTED");
        ESP_LOGE(TAG, "memory free : %d bytes", esp_get_free_heap_size());
        break;
    }
    return ESP_OK;
}

static void esp_ota_task(void *pvParameter)
{
    ESP_LOGI(TAG, "Starting OTA process !");
    esp_http_client_config_t config = {
        .event_handler = http_event_handler,
        .timeout_ms = 15 * 1000,
    };
    config.url = s_ota_url;
    // config.skip_cert_common_name_check = true;

    int result;

    if (s_sign_flag) {
        mbedtls_md_init(&md_ctx);
        result = mbedtls_md_setup(&md_ctx, mbedtls_md_info_from_type(MBEDTLS_MD_SHA256), 0);
        if (result != 0) {
            ESP_LOGE(TAG, "Setup md info failed.");
        }
        mbedtls_md_starts(&md_ctx);
    }

    esp_err_t ret = esp_https_ota(&config);
    if (ret == ESP_OK) {
        if (s_sign_flag) {
            unsigned char hash_buf[32];
            mbedtls_md_finish(&md_ctx, hash_buf);
            result = verify_ota_bin(hash_buf, 32, s_sign_data, s_sign_len);
            if (result != 0) {
                ESP_LOGE(TAG, "Verify ota bin signature failed : %d !", result);
                ESP_LOGE(TAG, "Firmware upgrade failed");
                ota_is_running = 0;

                report_device_soft_update_property_change(0,
                                                          SOFTWARE_UPDATE_UPDATE_STATE_IDLE,
                                                          SOFTWARE_UPDATE_UPDATE_ERROR_SIGNATURE_INVALID,
                                                          FIRMWARE_VERSION);
                esp_ota_mark_app_invalid_rollback_and_reboot();
            } else {
                ESP_LOGI(TAG, "Firmware upgrade success, now reboot !");
                esp_restart();
            }
        } else {
            ESP_LOGI(TAG, "Firmware upgrade success, now reboot !");
            esp_restart();
        }
    } else {
        ESP_LOGE(TAG, "Firmware upgrade failed");
        ota_is_running = 0;
        switch (ret) {
        case ESP_ERR_NO_MEM:
            report_device_soft_update_property_change(0,
                                                      SOFTWARE_UPDATE_UPDATE_STATE_IDLE,
                                                      SOFTWARE_UPDATE_UPDATE_ERROR_NOT_ENOUGH_RAM_MEMORY,
                                                      FIRMWARE_VERSION);
            break;
        case ESP_ERR_OTA_VALIDATE_FAILED:
            report_device_soft_update_property_change(0,
                                                      SOFTWARE_UPDATE_UPDATE_STATE_IDLE,
                                                      SOFTWARE_UPDATE_UPDATE_ERROR_INTEGRITY_CHECK_FAILURE,
                                                      FIRMWARE_VERSION);
            break;
        case ESP_ERR_INVALID_ARG:
            report_device_soft_update_property_change(0,
                                                      SOFTWARE_UPDATE_UPDATE_STATE_IDLE,
                                                      SOFTWARE_UPDATE_UPDATE_ERROR_INVALID_URL,
                                                      FIRMWARE_VERSION);
            break;
        case ESP_ERR_FLASH_OP_TIMEOUT:
        case ESP_ERR_FLASH_OP_FAIL:
            report_device_soft_update_property_change(0,
                                                      SOFTWARE_UPDATE_UPDATE_STATE_IDLE,
                                                      SOFTWARE_UPDATE_UPDATE_ERROR_UPDATE_FAILED,
                                                      FIRMWARE_VERSION);
            break;
        default:
            report_device_soft_update_property_change(0,
                                                      SOFTWARE_UPDATE_UPDATE_STATE_IDLE,
                                                      SOFTWARE_UPDATE_UPDATE_ERROR_CONNECTION_LOST,
                                                      FIRMWARE_VERSION);
            break;
        }
    }

    if (NULL != s_sign_data) {
        free(s_sign_data);
    }
    if (s_sign_flag) {
        mbedtls_md_free(&md_ctx);
    }

    vTaskDelete(NULL);
}

/**
 * @brief start a ota task with firmware bin file url
 * @param url       [firmware bin file url]
 * @param sign_flag [if firmware bin need signature verify]
 * @param char      [signature data, ASN.1 format]
 * @param sign_len  [signature data length]
 * @return int      [start result]
 */
int start_ota_with_url(char *url, bool sign_flag, unsigned char *sign, int sign_len)
{
    strcpy(s_ota_url, url);

    if (sign_flag) {
        s_sign_data = malloc(sign_len);
        if (NULL == s_sign_data) {
            ESP_LOGE(TAG, "Malloc memary for sign data failed.");
            return -1;
        }
        memcpy(s_sign_data, sign, sign_len);
        s_sign_len = sign_len;
        s_sign_flag = sign_flag;
    }

    if (ota_is_running == 0) {
        ota_is_running = 1;
        current_len = 0;
        xTaskCreate(&esp_ota_task, "ota_task", 4096, NULL, 5, NULL);
    } else {
        ESP_LOGE(TAG, "OTA process is running !");
        return -1;
    }

    return 0;
}
