/**
 * Copyright (c) 2008-2021, OPPO Mobile Comm Corp., Ltd
 * VENDOR_EDIT
 * @file light_service.c
 * @brief
 * @version
 * @date 2021-4
 * @author
 */

#include <errno.h>

#include "light.pb-c.h"
#include "light_service.h"
#include "hey_schema_api.h"

#define LIGHT_SERVICE 512
#define PROP_IID_CNT 14
#define ACT_IID_CNT 1

static const uint32_t light_properties_iids[] = { 1, 2, 3, 4, 5, 6, 7, 9, 11, 12, 13, 14, 15, 16 };
static struct hey_service *light_service;

static int action_light_language_call_cb(void *cb_data, struct hey_action_ctx *ctx, size_t len, unsigned char *value, void **user_context)
{
    int ret = -1;
    Iot__Service__Light__ActionInLightLanguage *light_language = iot__service__light__action_in_light_language__unpack(NULL, len, value);
    if (NULL == light_language) {
        hey_action_stop(ctx);
        return EINVAL;
    }

    struct light_light_language_in action;
    action.mode = light_language->mode;
    ret = light_light_language_call_cb(ctx, &action);

    iot__service__light__action_in_light_language__free_unpacked(light_language, NULL);
    return ret;
}

static int service_light_properties_write_cb(void *cb_data, size_t n_iid, unsigned int *iids,
                                             size_t len, unsigned char *value)
{
    Iot__Service__Light__Properties *props = iot__service__light__properties__unpack(NULL, len, value);
    if (NULL == props) {
        return EINVAL;
    }

    struct light_properties properties;
    struct light_rgb colorrgb;
    struct light_free_control_night_light freecontrolnightlight;
    struct hey_array_node *scenecttbrightness[props->n_scenecttbrightness];
    for (int i = 0; i < props->n_scenecttbrightness; i++) {
        scenecttbrightness[i]->id = props->scenecttbrightness[i]->id;
        scenecttbrightness[i]->data.data = props->scenecttbrightness[i]->data.data;
        scenecttbrightness[i]->data.len = props->scenecttbrightness[i]->data.len;
    }
    struct light_study_mode_setting studymodesetting;
    if (props->colorrgb) {
        colorrgb.r = props->colorrgb->r;
        colorrgb.g = props->colorrgb->g;
        colorrgb.b = props->colorrgb->b;
    }
    if (props->freecontrolnightlight) {
        freecontrolnightlight.enable = props->freecontrolnightlight->enable;
        freecontrolnightlight.starthour = props->freecontrolnightlight->starthour;
        freecontrolnightlight.startmin = props->freecontrolnightlight->startmin;
        freecontrolnightlight.endhour = props->freecontrolnightlight->endhour;
        freecontrolnightlight.endmin = props->freecontrolnightlight->endmin;
    }
    if (props->studymodesetting) {
        studymodesetting.repeatnum = props->studymodesetting->repeatnum;
        studymodesetting.studytime = props->studymodesetting->studytime;
        studymodesetting.breaktime = props->studymodesetting->breaktime;
    }
    properties.power = props->power;
    properties.brightness = props->brightness;
    properties.cct = props->cct;
    properties.lightmode = props->lightmode;
    properties.autocolortemperature = props->autocolortemperature;
    properties.colorrgb = &colorrgb;
    properties.colorsaturation = props->colorsaturation;
    properties.autobrightness = props->autobrightness;
    properties.freecontrolnightlight = &freecontrolnightlight;
    properties.n_scenecttbrightness = props->n_scenecttbrightness;
    properties.scenecttbrightness = scenecttbrightness;
    properties.colorcttmode = props->colorcttmode;
    properties.studymodesetting = &studymodesetting;
    properties.appointmentpoweron = props->appointmentpoweron;
    properties.appointmentpoweroff = props->appointmentpoweroff;
    int ret = light_properties_write_cb(n_iid, iids, &properties);

    iot__service__light__properties__free_unpacked(props, NULL);
    return ret;
}

static const struct hey_properties_callbacks props_cbs = {
    .write = service_light_properties_write_cb
};

static const struct hey_action_ctx_callbacks light_language_cbs = {
    .start = action_light_language_call_cb
};

static int light_service_register(Iot__Service__Light__Properties *props)
{
    unsigned char *buf;
    size_t buf_len;
    buf_len = iot__service__light__properties__get_packed_size(props);
    if (buf_len == 0) {
        buf = NULL;
    } else {
        buf = malloc(buf_len);
        if (!buf) {
            return ENOMEM;
        }
        iot__service__light__properties__pack(props, buf);
    }

    struct hey_action_entry *actions_iids[ACT_IID_CNT];
    uint8_t actions_n_iid = 0;

    struct hey_action_entry act_light_language;
    act_light_language.iid = LIGHT_ACTION_LIGHT_LANGUAGE;
    act_light_language.type = HEY_ACTION_ONCE;
    act_light_language.cbs = &light_language_cbs;
    actions_iids[actions_n_iid++] = &act_light_language;

    struct hey_schema_service schema_srv;
    hey_schema_service_init(&schema_srv);
    schema_srv.properties.iids = (uint32_t *)light_properties_iids;
    schema_srv.properties.n_iid = PROP_IID_CNT;
    schema_srv.properties.value = buf;
    schema_srv.properties.len = buf_len;
    schema_srv.properties.cbs = &props_cbs;
    schema_srv.actions.iids = *actions_iids;
    schema_srv.actions.n_iid = ACT_IID_CNT;
    light_service = hey_service_register(LIGHT_SERVICE, &schema_srv, NULL);
    if (NULL == light_service) {
        free(buf);
        return errno;
    }

    free(buf);
    return 0;
}

/**
 * @brief update properties of light service
 * @param count      [the number of properties to be updated]
 * @param enumerate  [the iid enumeration of properties to be updated]
 * @param properties [properties structure with updated content]
 * @return int       [update result]
 */
int light_properties_update(size_t count, unsigned int *enumerate, struct light_properties *properties)
{
    if (NULL == properties) {
        return EINVAL;
    }

    int ret = 0;
    Iot__Service__Light__Properties *props = NULL;
    unsigned char *buf = NULL;
    Iot__Service__Light__RGB *colorrgb = NULL;
    Iot__Service__Light__FreeControlNightLight *freecontrolnightlight = NULL;
    ArrayNode **scenecttbrightness = NULL;
    Iot__Service__Light__StudyModeSetting *studymodesetting = NULL;
    props = malloc(sizeof(Iot__Service__Light__Properties));
    if (NULL == props) {
        ret = ENOMEM;
        goto out;
    }
    iot__service__light__properties__init(props);
    if (properties->colorrgb) {
        colorrgb = malloc(sizeof(Iot__Service__Light__RGB));
        if (NULL == colorrgb) {
            ret = ENOMEM;
            goto out;
        }
        iot__service__light__rgb__init(colorrgb);
        colorrgb->r = properties->colorrgb->r;
        colorrgb->g = properties->colorrgb->g;
        colorrgb->b = properties->colorrgb->b;
    }
    if (properties->freecontrolnightlight) {
        freecontrolnightlight = malloc(sizeof(Iot__Service__Light__FreeControlNightLight));
        if (NULL == freecontrolnightlight) {
            ret = ENOMEM;
            goto out;
        }
        iot__service__light__free_control_night_light__init(freecontrolnightlight);
        freecontrolnightlight->enable = properties->freecontrolnightlight->enable;
        freecontrolnightlight->starthour = properties->freecontrolnightlight->starthour;
        freecontrolnightlight->startmin = properties->freecontrolnightlight->startmin;
        freecontrolnightlight->endhour = properties->freecontrolnightlight->endhour;
        freecontrolnightlight->endmin = properties->freecontrolnightlight->endmin;
    }
    if (properties->n_scenecttbrightness) {
        scenecttbrightness = malloc(properties->n_scenecttbrightness * sizeof(ArrayNode));
        if (NULL == scenecttbrightness) {
            ret = ENOMEM;
            goto out;
        }
        for (int i = 0; i < properties->n_scenecttbrightness; i++) {
            array_node__init(scenecttbrightness[i]);
            scenecttbrightness[i]->id = properties->scenecttbrightness[i]->id;
            scenecttbrightness[i]->data.data = properties->scenecttbrightness[i]->data.data;
            scenecttbrightness[i]->data.len = properties->scenecttbrightness[i]->data.len;
        }
    }
    if (properties->studymodesetting) {
        studymodesetting = malloc(sizeof(Iot__Service__Light__StudyModeSetting));
        if (NULL == studymodesetting) {
            ret = ENOMEM;
            goto out;
        }
        iot__service__light__study_mode_setting__init(studymodesetting);
        studymodesetting->repeatnum = properties->studymodesetting->repeatnum;
        studymodesetting->studytime = properties->studymodesetting->studytime;
        studymodesetting->breaktime = properties->studymodesetting->breaktime;
    }
    props->power = properties->power;
    props->brightness = properties->brightness;
    props->cct = properties->cct;
    props->lightmode = properties->lightmode;
    props->autocolortemperature = properties->autocolortemperature;
    props->colorrgb = colorrgb;
    props->colorsaturation = properties->colorsaturation;
    props->autobrightness = properties->autobrightness;
    props->freecontrolnightlight = freecontrolnightlight;
    props->n_scenecttbrightness = properties->n_scenecttbrightness;
    props->scenecttbrightness = scenecttbrightness;
    props->colorcttmode = properties->colorcttmode;
    props->studymodesetting = studymodesetting;
    props->appointmentpoweron = properties->appointmentpoweron;
    props->appointmentpoweroff = properties->appointmentpoweroff;

    size_t buf_len;
    buf_len = iot__service__light__properties__get_packed_size(props);
    if (buf_len > 0) {
        buf = malloc(buf_len);
        if (NULL == buf) {
            ret = ENOMEM;
            goto out;
        }
        iot__service__light__properties__pack(props, buf);
    }

    struct hey_properties_update_content content = {
        .n_iid = count,
        .iids = enumerate,
        .len = buf_len,
        .value = buf
    };
    ret = hey_properties_update(light_service, &content);

out:
    free(props);
    free(buf);
    free(colorrgb);
    free(freecontrolnightlight);
    free(scenecttbrightness);
    free(studymodesetting);
    return ret;
}

/**
 * @brief light service initialization
 * @param properties [light service properties]
 * @return int       [initialize result]
*/
int light_service_init(const struct light_properties *properties)
{
    int ret = 0;
    Iot__Service__Light__Properties *props = NULL;
    Iot__Service__Light__RGB *colorrgb = NULL;
    Iot__Service__Light__FreeControlNightLight *freecontrolnightlight = NULL;
    ArrayNode **scenecttbrightness = NULL;
    Iot__Service__Light__StudyModeSetting *studymodesetting = NULL;
    props = malloc(sizeof(Iot__Service__Light__Properties));
    if (NULL == props) {
        ret = ENOMEM;
        goto out;
    }
    iot__service__light__properties__init(props);
    if (NULL != properties) {
        if (properties->colorrgb) {
            colorrgb = malloc(sizeof(Iot__Service__Light__RGB));
            if (NULL == colorrgb) {
                ret = ENOMEM;
                goto out;
            }
            iot__service__light__rgb__init(colorrgb);
            colorrgb->r = properties->colorrgb->r;
            colorrgb->g = properties->colorrgb->g;
            colorrgb->b = properties->colorrgb->b;
        }
        if (properties->freecontrolnightlight) {
            freecontrolnightlight = malloc(sizeof(Iot__Service__Light__FreeControlNightLight));
            if (NULL == freecontrolnightlight) {
                ret = ENOMEM;
                goto out;
            }
            iot__service__light__free_control_night_light__init(freecontrolnightlight);
            freecontrolnightlight->enable = properties->freecontrolnightlight->enable;
            freecontrolnightlight->starthour = properties->freecontrolnightlight->starthour;
            freecontrolnightlight->startmin = properties->freecontrolnightlight->startmin;
            freecontrolnightlight->endhour = properties->freecontrolnightlight->endhour;
            freecontrolnightlight->endmin = properties->freecontrolnightlight->endmin;
        }
        if (properties->n_scenecttbrightness) {
            scenecttbrightness = malloc(properties->n_scenecttbrightness * sizeof(ArrayNode));
            if (NULL == scenecttbrightness) {
                ret = ENOMEM;
                goto out;
            }
            for (int i = 0; i < properties->n_scenecttbrightness; i++) {
                array_node__init(scenecttbrightness[i]);
                scenecttbrightness[i]->id = properties->scenecttbrightness[i]->id;
                scenecttbrightness[i]->data.data = properties->scenecttbrightness[i]->data.data;
                scenecttbrightness[i]->data.len = properties->scenecttbrightness[i]->data.len;
            }
        }
        if (properties->studymodesetting) {
            studymodesetting = malloc(sizeof(Iot__Service__Light__StudyModeSetting));
            if (NULL == studymodesetting) {
                ret = ENOMEM;
                goto out;
            }
            iot__service__light__study_mode_setting__init(studymodesetting);
            studymodesetting->repeatnum = properties->studymodesetting->repeatnum;
            studymodesetting->studytime = properties->studymodesetting->studytime;
            studymodesetting->breaktime = properties->studymodesetting->breaktime;
        }
        props->power = properties->power;
        props->brightness = properties->brightness;
        props->cct = properties->cct;
        props->lightmode = properties->lightmode;
        props->autocolortemperature = properties->autocolortemperature;
        props->colorrgb = colorrgb;
        props->colorsaturation = properties->colorsaturation;
        props->autobrightness = properties->autobrightness;
        props->freecontrolnightlight = freecontrolnightlight;
        props->n_scenecttbrightness = properties->n_scenecttbrightness;
        props->scenecttbrightness = scenecttbrightness;
        props->colorcttmode = properties->colorcttmode;
        props->studymodesetting = studymodesetting;
        props->appointmentpoweron = properties->appointmentpoweron;
        props->appointmentpoweroff = properties->appointmentpoweroff;
    }

    ret = light_service_register(props);

out:
    free(props);
    free(colorrgb);
    free(freecontrolnightlight);
    free(scenecttbrightness);
    free(studymodesetting);
    return ret;
}

/**
 * @brief light language action return
 * @param ctx    [action context]
 * @param action [light language action out]
 * @return int   [action return result, 0 means success]
 */
int light_light_language_return(struct hey_action_ctx *ctx, struct light_light_language_out *action)
{
    int ret = 0;
    ret = hey_action_msg_send(ctx, 0, NULL);
    hey_action_stop(ctx);
    return ret;
}
