/**
 * Copyright (c) 2008-2021, OPPO Mobile Comm Corp., Ltd
 * VENDOR_EDIT
 * @file software_update_service.c
 * @brief
 * @version
 * @date 2021-4
 * @author
 */

#include <errno.h>

#include "softwareUpdate.pb-c.h"
#include "software_update_service.h"
#include "hey_schema_api.h"

#define SOFTWARE_UPDATE_SERVICE 16128
#define PROP_IID_CNT 4
#define ACT_IID_CNT 3

static const uint32_t software_update_properties_iids[] = { 1, 2, 3, 4 };
static struct hey_service *software_update_service;

static int action_start_update_call_cb(void *cb_data, struct hey_action_ctx *ctx, size_t len, unsigned char *value, void **user_context)
{
    int ret = -1;
    Iot__Service__SoftwareUpdate__ActionInStartUpdate *start_update = iot__service__software_update__action_in_start_update__unpack(NULL, len, value);
    if (NULL == start_update) {
        hey_action_stop(ctx);
        return EINVAL;
    }

    struct software_update_start_update_in action;
    action.packageurl = start_update->packageurl;
    action.validatefromver = start_update->validatefromver;
    action.signtype = start_update->signtype;
    action.signatures.data = start_update->signatures.data;
    action.signatures.len = start_update->signatures.len;
    ret = software_update_start_update_call_cb(ctx, &action);

    iot__service__software_update__action_in_start_update__free_unpacked(start_update, NULL);
    return ret;
}

static int action_start_update_file_start_cb(void *cb_data, struct hey_action_ctx *ctx, size_t len, unsigned char *value, void **user_context)
{
    int ret = -1;
    Iot__Service__SoftwareUpdate__ActionInStartUpdateFile *start_update_file = iot__service__software_update__action_in_start_update_file__unpack(NULL, len, value);
    if (NULL == start_update_file) {
        hey_action_stop(ctx);
        return EINVAL;
    }

    struct software_update_start_update_file_in action;
    action.send_type_case = start_update_file->send_type_case;
    switch (action.send_type_case) {
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_NOT_SET:
        break;
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_FILEINFO: {
        struct software_update_start_update_file_file_info fileinfo;
        fileinfo.validatefromver = start_update_file->fileinfo->validatefromver;
        fileinfo.signtype = start_update_file->fileinfo->signtype;
        fileinfo.signatures.data = start_update_file->fileinfo->signatures.data;
        fileinfo.signatures.len = start_update_file->fileinfo->signatures.len;
        fileinfo.size = start_update_file->fileinfo->size;
        action.fileinfo = &fileinfo;
    } break;
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_CHUNK:
        action.chunk.data = start_update_file->chunk.data;
        action.chunk.len = start_update_file->chunk.len;
        break;
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_FINAL:
        break;
    }
    ret = software_update_start_update_file_start_cb(ctx, &action, user_context);

    iot__service__software_update__action_in_start_update_file__free_unpacked(start_update_file, NULL);
    return ret;
}

static int action_start_update_file_message_cb(void *cb_data, void *user_context, size_t len, unsigned char *value)
{
    int ret = -1;
    Iot__Service__SoftwareUpdate__ActionInStartUpdateFile *start_update_file = iot__service__software_update__action_in_start_update_file__unpack(NULL, len, value);
    if (NULL == start_update_file) {
        return EINVAL;
    }

    struct software_update_start_update_file_in action;
    action.send_type_case = start_update_file->send_type_case;
    switch (action.send_type_case) {
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_NOT_SET:
        break;
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_FILEINFO: {
        struct software_update_start_update_file_file_info fileinfo;
        fileinfo.validatefromver = start_update_file->fileinfo->validatefromver;
        fileinfo.signtype = start_update_file->fileinfo->signtype;
        fileinfo.signatures.data = start_update_file->fileinfo->signatures.data;
        fileinfo.signatures.len = start_update_file->fileinfo->signatures.len;
        fileinfo.size = start_update_file->fileinfo->size;
        action.fileinfo = &fileinfo;
    } break;
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_CHUNK:
        action.chunk.data = start_update_file->chunk.data;
        action.chunk.len = start_update_file->chunk.len;
        break;
    case SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_FINAL:
        break;
    }
    ret = software_update_start_update_file_message_cb(user_context, &action);

    iot__service__software_update__action_in_start_update_file__free_unpacked(start_update_file, NULL);
    return ret;
}

static void action_start_update_file_stop_cb(void *cb_data, void *user_context)
{
    software_update_start_update_file_stop_cb(user_context);
}

/**
 * @brief start update file action send message of software update service
 * @param ctx    [action context]
 * @param action [start update file action out]
 * @return int   [action send result, 0 means success]
 */
int software_update_start_update_file_send(struct hey_action_ctx *ctx, struct software_update_start_update_file_out *action)
{
    int ret = 0;
    Iot__Service__SoftwareUpdate__ActionOutStartUpdateFile *out = NULL;
    unsigned char *buf = NULL;
    out = malloc(sizeof(Iot__Service__SoftwareUpdate__ActionOutStartUpdateFile));
    if (NULL == out) {
        ret = ENOMEM;
        goto out;
    }
    iot__service__software_update__action_out_start_update_file__init(out);
    out->code = action->code;
    size_t buf_len = iot__service__software_update__action_out_start_update_file__get_packed_size(out);
    if (buf_len > 0) {
        buf = malloc(buf_len);
        if (NULL == buf) {
            ret = ENOMEM;
            goto out;
        }
    }
    iot__service__software_update__action_out_start_update_file__pack(out, buf);
    ret = hey_action_msg_send(ctx, buf_len, buf);

out:
    free(out);
    free(buf);
    return ret;
}

/**
 * @brief start update file action stop of software update service
 * @param ctx  [action context]
 * @return int [action stop result, 0 means success]
 */
int software_update_start_update_file_stop(struct hey_action_ctx *ctx)
{
    return hey_action_stop(ctx);
}

static int action_query_new_update_call_cb(void *cb_data, struct hey_action_ctx *ctx, size_t len, unsigned char *value, void **user_context)
{
    int ret = -1;
    Iot__Service__SoftwareUpdate__ActionInQueryNewUpdate *query_new_update = iot__service__software_update__action_in_query_new_update__unpack(NULL, len, value);
    if (NULL == query_new_update) {
        hey_action_stop(ctx);
        return EINVAL;
    }

    struct software_update_query_new_update_in action;
    ret = software_update_query_new_update_call_cb(ctx, &action);

    iot__service__software_update__action_in_query_new_update__free_unpacked(query_new_update, NULL);
    return ret;
}

static const struct hey_action_ctx_callbacks start_update_cbs = {
    .start = action_start_update_call_cb
};

static const struct hey_action_ctx_callbacks start_update_file_cbs = {
    .start = action_start_update_file_start_cb,
    .message = action_start_update_file_message_cb,
    .stop = action_start_update_file_stop_cb
};

static const struct hey_action_ctx_callbacks query_new_update_cbs = {
    .start = action_query_new_update_call_cb
};

static int software_update_service_register(Iot__Service__SoftwareUpdate__Properties *props)
{
    unsigned char *buf;
    size_t buf_len;
    buf_len = iot__service__software_update__properties__get_packed_size(props);
    if (buf_len == 0) {
        buf = NULL;
    } else {
        buf = malloc(buf_len);
        if (!buf) {
            return ENOMEM;
        }
        iot__service__software_update__properties__pack(props, buf);
    }

    struct hey_action_entry *actions_iids[ACT_IID_CNT];
    uint8_t actions_n_iid = 0;

    struct hey_action_entry act_start_update;
    act_start_update.iid = SOFTWARE_UPDATE_ACTION_START_UPDATE;
    act_start_update.type = HEY_ACTION_ONCE;
    act_start_update.cbs = &start_update_cbs;
    actions_iids[actions_n_iid++] = &act_start_update;

    struct hey_action_entry act_start_update_file;
    act_start_update_file.iid = SOFTWARE_UPDATE_ACTION_START_UPDATE_FILE;
    act_start_update_file.type = HEY_ACTION_IN_STREAM_OUT_ONCE;
    act_start_update_file.cbs = &start_update_file_cbs;
    actions_iids[actions_n_iid++] = &act_start_update_file;

    struct hey_action_entry act_query_new_update;
    act_query_new_update.iid = SOFTWARE_UPDATE_ACTION_QUERY_NEW_UPDATE;
    act_query_new_update.type = HEY_ACTION_ONCE;
    act_query_new_update.cbs = &query_new_update_cbs;
    actions_iids[actions_n_iid++] = &act_query_new_update;

    struct hey_schema_service schema_srv;
    hey_schema_service_init(&schema_srv);
    schema_srv.properties.iids = (uint32_t *)software_update_properties_iids;
    schema_srv.properties.n_iid = PROP_IID_CNT;
    schema_srv.properties.value = buf;
    schema_srv.properties.len = buf_len;
    schema_srv.actions.iids = *actions_iids;
    schema_srv.actions.n_iid = ACT_IID_CNT;
    software_update_service = hey_service_register(SOFTWARE_UPDATE_SERVICE, &schema_srv, NULL);
    if (NULL == software_update_service) {
        free(buf);
        return errno;
    }

    free(buf);
    return 0;
}

/**
 * @brief update properties of software update service
 * @param count      [the number of properties to be updated]
 * @param enumerate  [the iid enumeration of properties to be updated]
 * @param properties [properties structure with updated content]
 * @return int       [update result]
 */
int software_update_properties_update(size_t count, unsigned int *enumerate, struct software_update_properties *properties)
{
    if (NULL == properties) {
        return EINVAL;
    }

    int ret = 0;
    Iot__Service__SoftwareUpdate__Properties *props = NULL;
    unsigned char *buf = NULL;
    props = malloc(sizeof(Iot__Service__SoftwareUpdate__Properties));
    if (NULL == props) {
        ret = ENOMEM;
        goto out;
    }
    iot__service__software_update__properties__init(props);
    props->currentsoftwareversion = properties->currentsoftwareversion;
    props->updateprogress = properties->updateprogress;
    props->updatestate = properties->updatestate;
    props->updateerror = properties->updateerror;

    size_t buf_len;
    buf_len = iot__service__software_update__properties__get_packed_size(props);
    if (buf_len > 0) {
        buf = malloc(buf_len);
        if (NULL == buf) {
            ret = ENOMEM;
            goto out;
        }
        iot__service__software_update__properties__pack(props, buf);
    }

    struct hey_properties_update_content content = {
        .n_iid = count,
        .iids = enumerate,
        .len = buf_len,
        .value = buf
    };
    ret = hey_properties_update(software_update_service, &content);

out:
    free(props);
    free(buf);
    return ret;
}

/**
 * @brief software update service initialization
 * @param properties [software update service properties]
 * @return int       [initialize result]
*/
int software_update_service_init(const struct software_update_properties *properties)
{
    int ret = 0;
    Iot__Service__SoftwareUpdate__Properties *props = NULL;
    props = malloc(sizeof(Iot__Service__SoftwareUpdate__Properties));
    if (NULL == props) {
        ret = ENOMEM;
        goto out;
    }
    iot__service__software_update__properties__init(props);
    if (NULL != properties) {
        props->currentsoftwareversion = properties->currentsoftwareversion;
        props->updateprogress = properties->updateprogress;
        props->updatestate = properties->updatestate;
        props->updateerror = properties->updateerror;
    }

    ret = software_update_service_register(props);

out:
    free(props);
    return ret;
}

/**
 * @brief start update action return
 * @param ctx    [action context]
 * @param action [start update action out]
 * @return int   [action return result, 0 means success]
 */
int software_update_start_update_return(struct hey_action_ctx *ctx, struct software_update_start_update_out *action)
{
    int ret = 0;
    ret = hey_action_msg_send(ctx, 0, NULL);
    hey_action_stop(ctx);
    return ret;
}

/**
 * @brief query new update action return
 * @param ctx    [action context]
 * @param action [query new update action out]
 * @return int   [action return result, 0 means success]
 */
int software_update_query_new_update_return(struct hey_action_ctx *ctx, struct software_update_query_new_update_out *action)
{
    int ret = 0;
    Iot__Service__SoftwareUpdate__ActionOutQueryNewUpdate *out = NULL;
    unsigned char *buf = NULL;
    out = malloc(sizeof(Iot__Service__SoftwareUpdate__ActionOutQueryNewUpdate));
    if (NULL == out) {
        ret = ENOMEM;
        goto out;
    }
    iot__service__software_update__action_out_query_new_update__init(out);
    out->currentver = action->currentver;
    out->newver = action->newver;
    out->desc = action->desc;
    size_t buf_len = iot__service__software_update__action_out_query_new_update__get_packed_size(out);
    if (buf_len > 0) {
        buf = malloc(buf_len);
        if (NULL == buf) {
            ret = ENOMEM;
            goto out;
        }
        iot__service__software_update__action_out_query_new_update__pack(out, buf);
    }
    ret = hey_action_msg_send(ctx, buf_len, buf);
    hey_action_stop(ctx);
out:
    free(out);
    free(buf);
    return ret;
}
