#ifndef SOFTWARE_UPDATE_SERVICE_H
#define SOFTWARE_UPDATE_SERVICE_H

#include "hey_schema_common.h"

enum software_update_update_state {
    SOFTWARE_UPDATE_UPDATE_STATE_IDLE = 0,
    SOFTWARE_UPDATE_UPDATE_STATE_PREPARE = 1,
    SOFTWARE_UPDATE_UPDATE_STATE_DOWNLOADING = 2,
    SOFTWARE_UPDATE_UPDATE_STATE_DOWNLOADED = 3,
    SOFTWARE_UPDATE_UPDATE_STATE_UPDATING = 4
};

enum software_update_update_error {
    SOFTWARE_UPDATE_UPDATE_ERROR_INITIAL = 0,
    SOFTWARE_UPDATE_UPDATE_ERROR_NOT_ENOUGH_FLASH_MEMORY = 1,
    SOFTWARE_UPDATE_UPDATE_ERROR_NOT_ENOUGH_RAM_MEMORY = 2,
    SOFTWARE_UPDATE_UPDATE_ERROR_CONNECTION_LOST = 3,
    SOFTWARE_UPDATE_UPDATE_ERROR_INTEGRITY_CHECK_FAILURE = 4,
    SOFTWARE_UPDATE_UPDATE_ERROR_INVALID_URL = 5,
    SOFTWARE_UPDATE_UPDATE_ERROR_UPDATE_FAILED = 6,
    SOFTWARE_UPDATE_UPDATE_ERROR_CURRENT_VERSION_ERR = 7,
    SOFTWARE_UPDATE_UPDATE_ERROR_SIGNATURE_TYPE_ERR = 8,
    SOFTWARE_UPDATE_UPDATE_ERROR_SIGNATURE_INVALID = 9
};

enum software_update_signature_type {
    SOFTWARE_UPDATE_SIGNATURE_TYPE_None = 0,
    SOFTWARE_UPDATE_SIGNATURE_TYPE_EcdsaSecp256r1Sha256 = 1
};

struct software_update_start_update_in {
    char *packageurl;
    char *validatefromver;
    enum software_update_signature_type signtype;
    struct hey_binary_data signatures;
};

struct software_update_start_update_out {
    char __place_holder__; // make compiler happy
};

struct software_update_start_update_file_file_info {
    char *validatefromver;
    enum software_update_signature_type signtype;
    struct hey_binary_data signatures;
    int64_t size;
};

struct software_update_start_update_file_final {
    char __place_holder__; // make compiler happy
};

enum software_update_start_update_file_send_type {
    SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_NOT_SET = 0,
    SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_FILEINFO = 1,
    SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_CHUNK = 2,
    SOFTWARE_UPDATE_START_UPDATE_FILE_SEND_TYPE_FINAL = 3
};

struct software_update_start_update_file_in {
    enum software_update_start_update_file_send_type send_type_case;
    union {
        struct software_update_start_update_file_file_info *fileinfo;
        struct hey_binary_data chunk;
        struct software_update_start_update_file_final *final;
    };
};

struct software_update_start_update_file_out {
    uint32_t code;
};

struct software_update_query_new_update_in {
    char __place_holder__; // make compiler happy
};

struct software_update_query_new_update_out {
    char *currentver;
    char *newver;
    char *desc;
};

enum SOFTWARE_UPDATE_PROPERTY {
    SOFTWARE_UPDATE_PROPERTY_CURRENTSOFTWAREVERSION = 1,
    SOFTWARE_UPDATE_PROPERTY_UPDATEPROGRESS = 2,
    SOFTWARE_UPDATE_PROPERTY_UPDATESTATE = 3,
    SOFTWARE_UPDATE_PROPERTY_UPDATEERROR = 4
};

enum SOFTWARE_UPDATE_ACTION {
    SOFTWARE_UPDATE_ACTION_START_UPDATE = 5,
    SOFTWARE_UPDATE_ACTION_START_UPDATE_FILE = 6,
    SOFTWARE_UPDATE_ACTION_QUERY_NEW_UPDATE = 7
};

struct software_update_properties {
    char *currentsoftwareversion;
    uint32_t updateprogress;
    enum software_update_update_state updatestate;
    enum software_update_update_error updateerror;
};

/**
 * @brief software update service initialization
 * @param properties [software update service properties]
 * @return int       [initialize result]
*/
int software_update_service_init(const struct software_update_properties *properties);

/**
 * @brief update properties of software update service
 * @param count      [the number of properties to be updated]
 * @param enumerate  [the iid enumeration of properties to be updated]
 * @param properties [properties structure with updated content]
 * @return int       [update result]
 */
int software_update_properties_update(size_t count, unsigned int *enumerate, struct software_update_properties *properties);

/**
 * @brief start update action call callback of software update service
 * @param ctx    [action context]
 * @param action [start update action in]
 * @return int   [action call handle result, 0 means success]
 */
int software_update_start_update_call_cb(struct hey_action_ctx *ctx, struct software_update_start_update_in *action);

/**
 * @brief start update action return
 * @param ctx    [action context]
 * @param action [start update action out]
 * @return int   [action return result, 0 means success]
 */
int software_update_start_update_return(struct hey_action_ctx *ctx, struct software_update_start_update_out *action);

/**
 * @brief start update file action start callback of software update service
 * @param ctx          [action context]
 * @param action       [start update file action in]
 * @param user_context [user context record in SDK as needed, otherwise set it NULL]
 * @return int         [action start handle result, 0 means success]
 */
int software_update_start_update_file_start_cb(struct hey_action_ctx *ctx, struct software_update_start_update_file_in *action, void **user_context);

/**
 * @brief start update file action send message of software update service
 * @param ctx    [action context]
 * @param action [start update file action out]
 * @return int   [action send result, 0 means success]
 */
int software_update_start_update_file_send(struct hey_action_ctx *ctx, struct software_update_start_update_file_out *action);

/**
 * @brief start update file action message receive callback of software update service
 * @param user_context [callback data which form action start callback function]
 * @param action       [start update file action in]
 * @return int         [action message callback handle result, 0 means success]
 */
int software_update_start_update_file_message_cb(void *user_context, struct software_update_start_update_file_in *action);

/**
 * @brief start update file action stop of software update service
 * @param ctx  [action context]
 * @return int [action stop result, 0 means success]
 */
int software_update_start_update_file_stop(struct hey_action_ctx *ctx);

/**
 * @brief start update file action stop callback of software update service
 *        closed by the peer, user must to call action stop function
 * @param user_context [callback data which form action start callback function]
 * @return none
 */
void software_update_start_update_file_stop_cb(void *user_context);

/**
 * @brief query new update action call callback of software update service
 * @param ctx    [action context]
 * @param action [query new update action in]
 * @return int   [action call handle result, 0 means success]
 */
int software_update_query_new_update_call_cb(struct hey_action_ctx *ctx, struct software_update_query_new_update_in *action);

/**
 * @brief query new update action return
 * @param ctx    [action context]
 * @param action [query new update action out]
 * @return int   [action return result, 0 means success]
 */
int software_update_query_new_update_return(struct hey_action_ctx *ctx, struct software_update_query_new_update_out *action);

#endif /* SOFTWARE_UPDATE_SERVICE_H */
